import java.util.List;
import processing.serial.*;
import gab.opencv.*;
import org.poly2tri.triangulation.point.*;
import org.poly2tri.triangulation.util.*;
import org.poly2tri.transform.coordinate.*;
import org.poly2tri.triangulation.sets.*;
import org.poly2tri.triangulation.delaunay.sweep.*;
import org.poly2tri.triangulation.delaunay.*;
import org.poly2tri.geometry.primitives.*;
import org.poly2tri.triangulation.*;
import org.poly2tri.geometry.polygon.*;
import org.poly2tri.*;

int GSType = 1;
int SERIES_NUM = 4;
boolean inversePolar = true;
boolean isPortrait = true;
String[] boardName = {
  "32x16", "32x16", "32x24H", "32x24H", "32x24H", "42x32H", "42x24H", "42x24H"
};
int[] intplFactor  = {
  1, 1, 2, 2, 1, 2, 2, 1
};
int[] boardAmnt    = {
  1, 2, 1, 2, 2, 1, 2, 2
};
int [] boardR      = {
  32, 32, 32, 32, 32, 42, 42, 42
};
int [] boardC      = {
  16, 16, 12, 12, 12, 16, 16, 16
};

//Serial
Serial port = null;
Serial port2 = null;

boolean initialize1 = false;
boolean initialize2 = false;
boolean flipSide = false;
boolean drawSensors = true;
boolean drawFPS = false;
boolean drawTilt = false;
boolean approximate = true;
boolean showRaw = false;
boolean drawDT = true;
boolean showLayers = false;

//Const
int THLD = 5; 
int CONST_MAX  = 20;
float MAX_SENS = CONST_MAX;
float SENS     = MAX_SENS-5;
float low_thr = 30;
float high_thr = 60;
float x_interval = 3.766f/2;
float y_interval = 3.774f/2;
float x_padding = 5.0f;
float y_padding = 5.0f;
float info_padding = 16;
int approxInterval = 8;
int NOISE_THRESHOLD = drawDT ? 50/approxInterval : 50;

int THLD_L3 = 225;
int THLD_L2 = 145;
int THLD_L1 = 90;
int THLD_L2Hang = 15;

//SensorData
int boardNum = boardAmnt[GSType];
int singleR  = boardR[GSType];
int singleC  = boardC[GSType];
int totalR   = singleR;
int totalC   = singleC*boardNum*intplFactor[GSType];
float[][] initSensorVal = new float[totalR][totalC];
float[][] hallSensorVal = new float[totalR][totalC];

//SampleData
BicubicInterpolator   bcp = new BicubicInterpolator();
BiLinearInterpolator  blp = new BiLinearInterpolator();
int sampleRes = 8;
int sampleR   = totalR * (sampleRes);
int sampleC   = totalC * (sampleRes);
float[][] sampleMVal  = new float[sampleR][sampleC];

//Stylus
float maxM = -1; //M = magnitude
float minM = 1;

PImage lowMapN, highMapN, lowMapS, highMapS, rawMap, sensors, L1Map, L2Map, L3Map, L2HangMap;
OpenCV opencv;
ArrayList<PVector> L1centroids;

int screenWidth = int(x_padding*2+(float)((totalC-1)*sampleRes*x_interval));
int screenHeight = int(y_padding*2+(float)((totalR-1)*sampleRes*y_interval)+info_padding);
float imgWidth = (float)((totalC-1)*sampleRes*x_interval);
float imgHeight = (float)((totalR-1)*sampleRes*y_interval);
boolean doCalibration = true;

int t_end = millis();

void setup() {
  frameRate(200);
  size(screenWidth, screenHeight);
  textSize(16);
  stroke(0);
  ellipseMode(CENTER);

  // List all the available serial ports
  String[] serialList = Serial.list();
  println("Found " + serialList.length + " serial device(s).");
  if (serialList.length == 0)
    exit();
  for (int i = 0; i < serialList.length; i++)
    println("["+(i+1)+"] "+serialList[i]);
  if (boardNum > 1 && serialList.length > 1) {
    port2 = new Serial(this, serialList[serialList.length-1], 921600);  // make sure Arduino is talking serial at this baud rate
    port2.write("a");
  }
  port = new Serial(this, serialList[serialList.length-boardNum], 921600);  // make sure Arduino is talking serial at this baud rate
  port.write("a");
  rawMap = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  opencv = new OpenCV(this, rawMap);
  L1centroids = new ArrayList<PVector>();

  background(0);
  fill(255);
  noStroke();
  for (int i = 0 ; i < totalC*totalR; i++)
    ellipse(((int)((totalC-1)-i%totalC)*sampleRes)*x_interval+x_padding, ((int)(i/totalC)*sampleRes)*y_interval+y_padding, 3, 3);
  saveFrame("sensors.jpg");
  sensors = loadImage("sensors.jpg");
  t_end = millis();
}

void draw() {
  if (doCalibration) {
    initialize1 = false;
    initialize2 = false;
    doCalibration = false;
  }
  else {
    background(0);
    upSampling();
    updateClipping();
    if (showRaw) drawRawData();
    else image(sensors, 0, 0);
    drawTokens();
    drawInfo();
  }
  t_end = millis();
  // Receive message from serial port
  getSerialMsg();
}

void keyReleased() {
  if (key == 'f' || key == 'F') {
    doCalibration = true;
    flipSide = !flipSide;
  } else if(key == 'a' || key == 'A'){ //<>//
    if(approximate)
      NOISE_THRESHOLD *= approxInterval;
    else
      NOISE_THRESHOLD /= approxInterval;
    approximate = !approximate;
  } else if (key == 'r' || key == 'R')
    showRaw = !showRaw;
  else if (key == 't' || key == 'T'){
    drawDT = !drawDT;
    drawTilt = !drawTilt;
  } else if (key == 'z' || key == 'Z') {
    doCalibration = true;
  } else if (key == 'l' || key == 'L')
    showLayers = !showLayers;
}
