void recordSampleData(boolean isBoundary, int i, int j){
  float sampleX, sampleY, sampleM;
  int w,h;
  for (int x=0; x<sampleRes; x++) {
    for (int y=0; y<sampleRes; y++) {
      sampleX = (float)x / (float)sampleRes;
      sampleY = (float)y / (float)sampleRes;
      w = i*sampleRes + x;
      h = j*sampleRes + y;
      if(isBoundary){
        sampleM = blp.getValue(sampleX, sampleY);
      }else{
        sampleM = bcp.getValue(sampleX, sampleY);
      }
      sampleMVal[w][h] = sampleM;    
    }
  }
}

void upSampling() {
  float[][] pBC = new float[4][4];
  float[][] pBL = new float[2][2];
  //BiLinear Interpolate the left and right edge
  for (int i=0;i<totalR;i+=totalR-2) {
    for (int j=1;j<totalC-2;j++) {
      pBL[0][0] = hallSensorVal[i  ][j];  
      pBL[0][1] = hallSensorVal[i  ][j+1];
      pBL[1][0] = hallSensorVal[i+1][j];  
      pBL[1][1] = hallSensorVal[i+1][j+1];   
      blp.updateCoefficients(pBL);
      recordSampleData(true,i,j); //true: isBoundary
    }
  }
  //BiLinear Interpolate the top and bottom edge
  for (int i=0;i<totalR-1;i++) {
    for (int j=0;j<totalC;j+=totalC-2) {
      pBL[0][0] = hallSensorVal[i  ][j];  
      pBL[0][1] = hallSensorVal[i  ][j+1];   
      pBL[1][0] = hallSensorVal[i+1][j];  
      pBL[1][1] = hallSensorVal[i+1][j+1]; 
      blp.updateCoefficients(pBL);
      recordSampleData(true,i,j); //true: isBoundary
    }
  }
  //Bicubic Interpolate other points
  for (int i=1;i<totalR-2;i++) {
    for (int j=1;j<totalC-2;j++) {
      pBC[0][0] = hallSensorVal[(i-1)][j-1];  
      pBC[0][1] = hallSensorVal[(i-1)][j];  
      pBC[0][2] = hallSensorVal[(i-1)][j+1];  
      pBC[0][3] = hallSensorVal[(i-1)][j+2];
      pBC[1][0] = hallSensorVal[(i)  ][j-1];  
      pBC[1][1] = hallSensorVal[(i)  ][j];  
      pBC[1][2] = hallSensorVal[(i)  ][j+1];  
      pBC[1][3] = hallSensorVal[(i)  ][j+2];
      pBC[2][0] = hallSensorVal[(i+1)][j-1];  
      pBC[2][1] = hallSensorVal[(i+1)][j];  
      pBC[2][2] = hallSensorVal[(i+1)][j+1];  
      pBC[2][3] = hallSensorVal[(i+1)][j+2]; 
      pBC[3][0] = hallSensorVal[(i+2)][j-1];  
      pBC[3][1] = hallSensorVal[(i+2)][j];  
      pBC[3][2] = hallSensorVal[(i+2)][j+1];  
      pBC[3][3] = hallSensorVal[(i+2)][j+2];
      bcp.updateCoefficients(pBC);
      recordSampleData(false,i,j); //false: isBoundary
    }
  }
}

void updateClipping(){
  int w,h;
  float pM;
  maxM = -1;
  minM = 1;
  rawMap = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  lowMapN = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  highMapN = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  lowMapS = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  highMapS = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  L2HangMap = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  L1Map = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  L2Map = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  L3Map = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  rawMap.loadPixels();
  lowMapN.loadPixels();
  highMapN.loadPixels();
  lowMapS.loadPixels();
  highMapS.loadPixels();
  L2HangMap.loadPixels();
  L1Map.loadPixels();
  L2Map.loadPixels();
  L3Map.loadPixels();

  for (int i=0;i<totalR-1;i++) {
    for (int j=0;j<totalC-1;j++) {
      for (int x=0; x<sampleRes; x++) {
        for (int y=0; y<sampleRes; y++) {
          w = i*sampleRes + x;
          h = j*sampleRes + y;
          pM = sampleMVal[w][h];
          if (pM>255) {
            pM = 255;
          }
          if (pM<-255) {
            pM = -255;
          }
          if(pM >= 0){    // N pole
            if (pM > maxM) {
              maxM = pM;
            }
            if (pM > high_thr){    // high threshold
              highMapN.pixels[h+w*highMapN.width] = color(0, 0, 255);
            }
            if (pM > low_thr){     // low threshold
              lowMapN.pixels[h+w*lowMapN.width] = color(0, 0, 255);
            }
            if (pM >= THLD_L1)
              L1Map.pixels[h+w*lowMapN.width] = color(0, 0, 255);
            else if (pM >= THLD_L2Hang)
              L2HangMap.pixels[h+w*lowMapN.width] = color(0, 0, 255);
            if (pM >= THLD_L2)
              L2Map.pixels[h+w*lowMapN.width] = color(0, 0, 255);
            if (pM >= THLD_L3)
              L3Map.pixels[h+w*lowMapN.width] = color(0, 0, 255);
            if(showRaw)rawMap.pixels[h+w*rawMap.width] = color(pM, 0, 0);
          } else{         // S pole
            if (pM < minM) {
              minM = pM;
            }
            if (pM < -high_thr){ // high threshold
              highMapS.pixels[h+w*highMapS.width] = color(255, 0, 0);
            }
            if (pM < -low_thr){  // low threshold
              lowMapS.pixels[h+w*lowMapS.width] = color(255, 0, 0);
            }
            if(showRaw)rawMap.pixels[h+w*rawMap.width] = color(0, 0, -pM);
          }
        }
      }
    }
  }
  rawMap.updatePixels();
  lowMapN.updatePixels();
  highMapN.updatePixels();
  lowMapS.updatePixels();
  highMapS.updatePixels();
  L2HangMap.updatePixels();
  L1Map.updatePixels();
  L2Map.updatePixels();
  L3Map.updatePixels();
}

void drawRawData() {
  image(rawMap, x_padding, y_padding, imgWidth, imgHeight);
  stroke(255);
  noFill();
  rect(x_padding, y_padding, imgWidth, imgHeight);
}

void drawTokens(){
  ArrayList<Contour> L1contoursN = new ArrayList<Contour>();
  ArrayList<Contour> L2contoursN = new ArrayList<Contour>();
  ArrayList<Contour> L1contoursS = new ArrayList<Contour>();
  ArrayList<Contour> L2contoursS = new ArrayList<Contour>();
  ArrayList<Contour> L1contours = new ArrayList<Contour>();
  ArrayList<Contour> L2contours = new ArrayList<Contour>();
  ArrayList<Contour> L3contours = new ArrayList<Contour>();
  ArrayList<Contour> L2Hangcontours = new ArrayList<Contour>();

  opencv.loadImage(lowMapN);
  L1contoursN = opencv.findContours();
  opencv.loadImage(highMapN);
  L2contoursN = opencv.findContours();
  opencv.loadImage(lowMapS);
  L1contoursS = opencv.findContours();
  opencv.loadImage(highMapS);
  L2contoursS = opencv.findContours();
  
  noFill();
  
  if(showLayers){
    opencv.loadImage(L2HangMap);
    for(int i = 0; i < 6; i++)
      opencv.erode();
    for(int i = 0; i < 5; i++)
      opencv.dilate();
    L2Hangcontours = opencv.findContours();
    opencv.loadImage(L1Map);
    for(int i = 0; i < 7; i++)
      opencv.dilate();
    for(int i = 0; i < 5; i++)
      opencv.erode();
    L1contours = opencv.findContours();
    opencv.loadImage(L2Map);
//    for(int i = 0; i < 5; i++)
//      opencv.dilate();
    L2contours = opencv.findContours();
    opencv.loadImage(L3Map);
//    for(int i = 0; i < 6; i++)
//      opencv.dilate();
    L3contours = opencv.findContours();
    
    strokeWeight(4);
    drawDT = false;
    approximate = false;
    drawContours(L1contours, false, color(255, 0, 0));
    drawContours(L2Hangcontours, false, color(100, 100, 255));
    drawContours(L2contours, false, color(100, 255, 100));
    drawContours(L3contours, false, color(255, 255, 0));
    drawDT = true;
    approximate = true;
  } else{
    strokeWeight(2);
    L1centroids.clear();
    drawContours(L1contoursN, true, color(128, 0, 0));
    drawContours(L2contoursN, false, color(255, 0, 0));
    drawContours(L1contoursS, !drawDT, color(0, 0, 128));
    drawContours(L2contoursS, false, color(0, 0, 255));
  }
}

void drawContours(ArrayList<Contour> s, boolean L1, color c){
  for (Contour contour : s) {
    ArrayList<PVector> points;
    points = contour.getPoints();
    if(approximate){
      for (int i = points.size()-1; i > 0; i--) {
        if(i % approxInterval != 0)
          points.remove(i);
      }
    }
    if (!drawDT && points.size() > NOISE_THRESHOLD) {
      stroke(c);
      beginShape();
      PVector centroid = new PVector(0, 0);
      PVector startPoint = null;
      for (PVector point : points) {
        if (startPoint == null)
          startPoint = point;
        vertex(point.x*x_interval+x_padding, point.y*y_interval+y_padding);
        centroid.x += point.x;
        centroid.y += point.y;
      }
      centroid.x /= points.size();
      centroid.y /= points.size();
      vertex(startPoint.x*x_interval+x_padding, startPoint.y*y_interval+y_padding);
      endShape();
      // Draw tilt
      if (drawTilt) {
        if (L1)
          L1centroids.add(centroid);
        else {
          float minDis = 10000;
          PVector pairCentroid = new PVector(0, 0);
          for (PVector L1centroid : L1centroids) {
            float dis = dist(L1centroid.x, L1centroid.y, centroid.x, centroid.y);
            if (dis < minDis) {
              minDis = dis;
              pairCentroid = L1centroid;
            }
          }
          stroke(0, 255, 0);
          line(pairCentroid.x*x_interval+x_padding, pairCentroid.y*y_interval+y_padding, 
          centroid.x*x_interval+x_padding, centroid.y*y_interval+y_padding);
          ellipse(pairCentroid.x*x_interval+x_padding, pairCentroid.y*y_interval+y_padding, 3, 3);
        }
      }
    } else if(L1 && drawDT && points.size() > NOISE_THRESHOLD)
      drawTri(points);
  }
}

void drawTri(ArrayList<PVector> points){
  // Delaunay Triangulation
  ArrayList<PolygonPoint> pointsList = new ArrayList<PolygonPoint>();
  for(PVector p : points)
    pointsList.add(new PolygonPoint(p.x*x_interval+x_padding, p.y*y_interval+y_padding));
  List<DelaunayTriangle> triangles;
  Polygon polygon = new Polygon(pointsList);
  Poly2Tri.triangulate(polygon);
  triangles = polygon.getTriangles();
  
  // Find & Trim Spine
  ArrayList<PVector> spineVertices = new ArrayList<PVector>();
  ArrayList<PVector> degree3Vertices = new ArrayList<PVector>();
  IntList degree4Index = new IntList();
  ArrayList<DelaunayTriangle> handledTri = new ArrayList<DelaunayTriangle>();
  strokeWeight(2);
  for (DelaunayTriangle triangle : triangles) {
    float fx = 0, fy = 0;
    float px = 0, py = 0;
    TriangulationPoint fp = triangle.points[0];
    TriangulationPoint pp = triangle.points[0];
    boolean bFirst = true;
    ArrayList<PVector> midPoints = new ArrayList<PVector>();
    for (TriangulationPoint point : triangle.points) {
      float x = point.getXf();
      float y = point.getYf();
      if (bFirst) {
        fx = x; fy = y;
        fp = point;
        bFirst = false;
      } else{
        if(triangle.cEdge[triangle.edgeIndex(pp, point)])  // constrained edge
          stroke(#FF0000);
        else{
          stroke(#005500);
          midPoints.add(new PVector((px+x)/2, (py+y)/2));
        }
        line(px, py, x, y);
      }
      px = x; py = y;
      pp = point;
    }
    if(triangle.cEdge[triangle.edgeIndex(pp, fp)])  // constrained edge
      stroke(#FF0000);
    else {
      stroke(#005500);
      midPoints.add(new PVector((px+fx)/2, (py+fy)/2));
    }
    line(px, py, fx, fy);
    if(midPoints.size() == 2) {
      spineVertices.add(new PVector(midPoints.get(0).x, midPoints.get(0).y));
      spineVertices.add(new PVector(midPoints.get(1).x, midPoints.get(1).y));
    } else if(midPoints.size() == 3) {
      DelaunayTriangle adjIntTri = null;
      for(int i = 0; i < 3; i++) {
        DelaunayTriangle neighborTri = triangle.neighbors[i];
        if(neighborTri != null)
          if(!neighborTri.cEdge[0] && !neighborTri.cEdge[1] && !neighborTri.cEdge[2]){
            adjIntTri = triangle.neighbors[i];
            break;
          }
      }
      float midx = (midPoints.get(0).x + midPoints.get(1).x + midPoints.get(2).x)/3;
      float midy = (midPoints.get(0).y + midPoints.get(1).y + midPoints.get(2).y)/3;
      if(adjIntTri == null || (adjIntTri != null && handledTri.contains(adjIntTri))){
        degree3Vertices.add(new PVector(midx, midy));
        for(int i = 0; i < midPoints.size(); i++){
          spineVertices.add(new PVector(midx, midy));
          spineVertices.add(new PVector(midPoints.get(i).x, midPoints.get(i).y));
        }
      } else{
        if(!handledTri.contains(triangle)){
          float adjMidx = (adjIntTri.points[0].getXf()+adjIntTri.points[1].getXf()+adjIntTri.points[2].getXf())/3;
          float adjMidy = (adjIntTri.points[0].getYf()+adjIntTri.points[1].getYf()+adjIntTri.points[2].getYf())/3;
          float midMidx = (midx+adjMidx)/2;
          float midMidy = (midy+adjMidy)/2;
          degree4Index.append(degree3Vertices.size());
          degree3Vertices.add(new PVector(midMidx, midMidy));
          PVector[] adjMidPoints = new PVector[3];
          adjMidPoints[0] = new PVector((adjIntTri.points[0].getXf()+adjIntTri.points[1].getXf())/2,(adjIntTri.points[0].getYf()+adjIntTri.points[1].getYf())/2);
          adjMidPoints[1] = new PVector((adjIntTri.points[2].getXf()+adjIntTri.points[1].getXf())/2,(adjIntTri.points[2].getYf()+adjIntTri.points[1].getYf())/2);
          adjMidPoints[2] = new PVector((adjIntTri.points[0].getXf()+adjIntTri.points[2].getXf())/2,(adjIntTri.points[0].getYf()+adjIntTri.points[2].getYf())/2);
          for(int i = 0; i < 3; i++){
            if(midPoints.contains(adjMidPoints[i]))
              midPoints.remove(adjMidPoints[i]);
            else
              midPoints.add(adjMidPoints[i]);
          }
          for(int i = 0; i < midPoints.size(); i++){
            spineVertices.add(new PVector(midMidx, midMidy));
            spineVertices.add(new PVector(midPoints.get(i).x, midPoints.get(i).y));
          }
          handledTri.add(adjIntTri);
        }
      }
    }
  }
  
  // Draw Simplified Spine
  stroke(#FFFF00);
  if(degree3Vertices.size() == 0 && spineVertices.size() > 1)
    line(spineVertices.get(0).x, spineVertices.get(0).y, 
         spineVertices.get(spineVertices.size()-1).x, spineVertices.get(spineVertices.size()-1).y);
  for(int i = 0; i < degree3Vertices.size(); i++){
    int degree = 3;
    for(int a = 0; a < degree4Index.size(); a++)
      if(i == degree4Index.get(a)){
        degree = 4;
        break;
      }
    for(int a = 0; a < degree; a++){
      PVector startp = degree3Vertices.get(i);
      PVector target = degree3Vertices.get(i);
      while(true){
        Boolean found = false;
        for(int j = 0; j < spineVertices.size(); j+=2){
          if(target.dist(spineVertices.get(j)) == 0){
            target = spineVertices.get(j+1);
            found = true;
          } else if(target.dist(spineVertices.get(j+1)) == 0){
            target = spineVertices.get(j);
            found = true;
          }
          if(found){
            spineVertices.remove(j+1);
            spineVertices.remove(j);
            for(int b = i+1; b < degree3Vertices.size(); b++)
              if(target.dist(degree3Vertices.get(b)) == 0){
                found = false;
                line(startp.x, startp.y, target.x, target.y);
                break;
              }
            break;
          }
        }
        if(!found)
          break;
        else if(startp.dist(target) > 70){
          line(startp.x, startp.y, target.x, target.y);
          startp = target;
        }
      }
      if(startp.dist(target) > 30)
        line(startp.x, startp.y, target.x, target.y);
    }
  }
}

void drawInfo() {
  float fRate = 1000.f/(millis()-t_end);
  if (fRate>30) 
    fill(0, 255, 0);
  else if (fRate>=15 && fRate<30)
    fill(255, 255, 0);
  else 
    fill(255, 0, 0);
  text("FPS: "+(int)fRate, 10, screenHeight);
  fill(255);
  text("MaxN: "+(int)maxM+" | MaxS: "+(int)minM, 100, screenHeight);
  if(doCalibration) 
    text("Calibrating...(Please Remove All Tokens from the Sensor)", 300, screenHeight);
  else 
    text("Calibrated.", 300, screenHeight);
}
